package ch.inf3.horizoncut.cli;

import ch.inf3.horizoncut.data.DisplayCalculator;
import ch.inf3.horizoncut.data.DistanceLayer;
import ch.inf3.horizoncut.data.GeoJsonExport;
import ch.inf3.horizoncut.data.Peak;
import ch.inf3.horizoncut.data.Position;
import ch.inf3.horizoncut.data.Tile;
import ch.inf3.horizoncut.data.TileMap;
import ch.inf3.horizoncut.data.TileReader;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.stream.Collectors;
import javax.imageio.ImageIO;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class Cli {

    private static final Logger LOG = LoggerFactory.getLogger(Cli.class);

    private static final int SLICE_DISTANCE = 100;

    public static void main(String[] args) throws Exception {

        Position c = new Position(46.326570, 7.466880);
        TileMap tileMap = new TileReader("/home/valdor/Downloads/SRTM1/").readHGTFiles(c, 2);

        final int width = 8000;
        final int height = 2000;

        Tile atOrigin = tileMap.getByLatLon(c.gridLat, c.gridLon);
        double eyeHeight = atOrigin.elevation(c.row, c.col);
        LOG.info("Eye at {} is at {}", c.getLatLon(), eyeHeight);

        double minHorizAngle = 135; //107; // 135 ~Brunegghorn
        double maxHorizAngle = 226; //226; // 197 ~Combin de Valsorey

        DisplayCalculator dc = new DisplayCalculator(width, height, minHorizAngle, maxHorizAngle, eyeHeight);

        int maxDistance = 100_000;
        int minDistance = 500;

        LOG.info("Distances of {} to {}", minDistance, maxDistance);

        BufferedImage img = new BufferedImage(width, height + 1, BufferedImage.TYPE_INT_RGB);

        Collection<DistanceLayer> layers = new ArrayList<>(maxDistance);
        for (int d = maxDistance; d > minDistance; d -= SLICE_DISTANCE) {
            final int cDist = d;
            double[] data = computeAtDistance(c, tileMap, cDist, dc);
            layers.add(new DistanceLayer(d - SLICE_DISTANCE, d, data));
        }

        // We compute the horizon and delete any layer that is behind the horizon
        // and is not used at all for the landscape.
        // For this, we are using the fact that we computed the layers starting
        // with the farthest ones in the build loop.
        DistanceLayer horizonLayer = layers.stream()
                .reduce((a, b) -> a.mergeWith(b))
                .get();

        layers = layers.stream()
                .dropWhile(l -> horizonLayer.isLower(l))
                .collect(Collectors.toList());

        LOG.info("There are {} layers remaining after horizon cleanup", layers.size());

        // We recompute the min and max distances after the cleanup of layers that
        // are not useful for the display.
        maxDistance = layers.stream()
                .map(l -> l.farDistance)
                .max(Integer::compareTo)
                .get();

        minDistance = layers.stream()
                .map(l -> l.nearDistance)
                .min(Integer::compareTo)
                .get();

        LOG.info("Distances of {} to {}", minDistance, maxDistance);

        Collection<DistanceLayer> nLayers = layers.stream()
                .collect(Collectors.groupingBy(l -> l.nearDistance / 1, Collectors.reducing((a, b) -> a.mergeWith(b))))
                .values().stream()
                .flatMap(o -> o.stream())
                .sorted((a, b) -> b.compareTo(a))
                .collect(Collectors.toList());

        nLayers = DistanceLayer.filterInvisible(nLayers);

        LOG.info("There are {} layers remaining after layer merges", nLayers.size());

        // We dont want to compte the min angle to avoid looking at the ground
        // just before the position ...
        double minVertAngle = 80;
        double maxVertAngle = nLayers.stream()
                .map(l -> l.getMax())
                .max(Double::compareTo)
                .get();

        maxVertAngle += 2;

        LOG.info("Displaying angles are {} - {}", minVertAngle, maxVertAngle);

        // We loop in descending order to avoid writing further points on top
        // of nearer points.
        for (DistanceLayer layer : nLayers) {
            float dPos = (float) ((maxDistance - layer.nearDistance) / (maxDistance * 1.0f));
            int color = Color.HSBtoRGB(dPos * 0.5f, 1.0f, 1.0f);

            for (int x = 0; x < width; x++) {
                double visibleAngle = layer.data[x];
                double pos = (visibleAngle - minVertAngle) * (height / (maxVertAngle - minVertAngle));

                int y = (int) (height - pos - 1);

                if (y <= 0 || y > height) {
                    y = height;
                }

                img.setRGB(x, y, color);
            }
        }

        Graphics2D g = (Graphics2D) img.getGraphics();
        Font font = new Font("Arial", Font.BOLD, 12);
        g.setColor(Color.white);
        g.setFont(font);

        Collection<Peak> peaks = readFeatures("data/osm-overpass-peaks.geojson");
        Collection<Peak> villages = readFeatures("data/osm-overpass-villages.geojson");
        for (int i = 0; i < 2; i++) {
            Collection<Peak> feature;
            int txtPosition;
            boolean displayInvisible;

            if (i == 0) {
                feature = peaks;
                txtPosition = 150;
                displayInvisible = false;
            } else {
                feature = villages;
                txtPosition = height - 50;
                displayInvisible = true;
            }

            for (Peak peak : feature) {
                double peakBearing = Math.toDegrees(Position.bearing(c, peak.position));
                if (!dc.isVisible(peakBearing)) {
                    LOG.info("Peak {} is not visible at bearing {}", peak.name, peakBearing);
                    continue;
                }

                int peakDistance = c.distanceToMeters(peak.position);

                Tile atPos = tileMap.getByLatLon(peak.position.gridLat, peak.position.gridLon);
                double elevation = atPos.elevation(peak.position.row, peak.position.col);

                double visibleAngle = dc.calculateVisibilityAngle(elevation, peakDistance);

                int x = dc.getXAtBearing(peakBearing);

                boolean isVisible = DistanceLayer.isVisible(layers, peakDistance, x, visibleAngle);
                if (!isVisible && !displayInvisible) {
                    LOG.info("Peak {} is not visible, hidden behind terrain.", peak.name);
                    continue;
                }

                if (visibleAngle < minVertAngle) {
                    LOG.info("Peak {} is not visible, too low.", peak.name);
                    continue;
                }

                double pos = (visibleAngle - minVertAngle) * (height / (maxVertAngle - minVertAngle));
                int y = (int) (height - pos - 1);

                LOG.info("Peak {} is at bearing {} ({}x{})", peak.name, peakBearing, x, y);

                String title = String.format("%s", peak.name);

                AffineTransform old = g.getTransform();
                g.rotate(Math.toRadians(-45), x, txtPosition);
                g.drawString(title, x, txtPosition);
                g.setTransform(old);

                g.drawLine(x, y, x, txtPosition);
            }
        }

        for (int i = 0; i < 360; i += 5) {
            int x = dc.getXAtBearing(i);

            g.drawLine(x, 0, x, height);
            g.drawString(i + "", x, 20);
        }

        g.dispose();

        try {
            String imageName = String.format("image-%05d.png", 1);
            File outputfile = new File("/home/valdor/", imageName);
            outputfile.createNewFile();

            ImageIO.write(img, "png", outputfile);
        } catch (IOException ex) {
            System.out.println("Exception while writing image: " + ex);
        }
    }

    static final double[] computeAtDistance(Position eyeCoordinates, TileMap tileMap,
            int d, DisplayCalculator dc) {

        double[] output = new double[dc.getWidth()];

        for (int x = 0; x < dc.getWidth(); x += 1) {
            double bearing = dc.getBearingAtX(x);
            Position target = Position.get(eyeCoordinates, Math.toRadians(bearing), d);

            Tile atPos = tileMap.getByLatLon(target.gridLat, target.gridLon);
            double elevation = atPos.elevation(target.row, target.col);

            int dMeters = eyeCoordinates.distanceToMeters(target);
            double visibleAngle = dc.calculateVisibilityAngle(elevation, dMeters);

            output[x] = visibleAngle;
        }

        return output;
    }

    public static Collection<Peak> readFeatures(String fileName) throws IOException {
        Collection<Peak> peaks = new ArrayList<>();

        try ( FileReader fr = new FileReader(fileName)) {
            Gson gson = new GsonBuilder().create();
            GeoJsonExport export = gson.fromJson(fr, GeoJsonExport.class);

            for (GeoJsonExport.Feature feature : export.features) {
                double[] coordinates = feature.geometry.coordinates;
                String name = feature.properties.name;
                Integer elevation = 0; // Integer.parseInt(feature.properties.ele);

                if (name == null || name.isBlank()) {
                    continue;
                }

                Peak peak = new Peak(coordinates[1], coordinates[0], name, elevation);
                peaks.add(peak);
            }
        }

        return peaks;
    }

}
