package ch.inf3.horizoncut.data;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;

public class DistanceLayer implements Comparable<DistanceLayer>, Cloneable {

    public final int nearDistance;
    public final int farDistance;
    public final double[] data;

    public DistanceLayer(int nearDistance, int farDistance, double[] data) {
        this.nearDistance = nearDistance;
        this.farDistance = farDistance;
        this.data = data;
    }

    @Override
    public int compareTo(DistanceLayer other) {
        return Integer.compare(this.nearDistance, other.nearDistance);
    }

    @Override
    protected DistanceLayer clone() {
        double[] cloneData = Arrays.copyOf(data, data.length);
        return new DistanceLayer(nearDistance, farDistance, cloneData);
    }

    public DistanceLayer mergeWith(DistanceLayer other) {
        DistanceLayer near = this.compareTo(other) <= 0 ? this : other;
        DistanceLayer far = this.compareTo(other) > 0 ? this : other;

        double[] output = new double[near.data.length];

        boolean anyNear = false;

        for (int d = 0; d < near.data.length; d += 1) {
            double yNear = near.data[d];
            double yFar = far.data[d];

            if (yNear > yFar) {
                output[d] = yNear;
                anyNear = true;
            } else {
                output[d] = yFar;
            }
        }

        int outputNearDistance = anyNear ? near.nearDistance : far.nearDistance;

        return new DistanceLayer(outputNearDistance, far.farDistance, output);
    }
    
    /**
     * Returns true if the given layer is always lower than the current layer.
     *
     * @param other
     * @return
     */
    public boolean isLower(DistanceLayer other) {
        for (int d = 0; d < this.data.length; d += 1) {
            double yThis = this.data[d];
            double yOther = other.data[d];

            if (yThis == 0 || yOther == 0) {
                continue;
            }

            if (yOther >= yThis) {
                return false;
            }
        }

        return true;
    }

    public double getMax() {
        return Arrays.stream(this.data)
                .boxed()
                .filter(e -> !e.isNaN())
                .max(Double::compareTo)
                .orElse(0.0);
    }
    
    public static Collection<DistanceLayer> filterInvisible(Collection<DistanceLayer> input) {
        
        List<DistanceLayer> sorted = input.stream()
                .sorted()
                .map(d -> d.clone())
                .collect(Collectors.toList());
        
        int layerSize = sorted.get(0).data.length;
        
        for (int x = 0; x < layerSize; x++) {
            double maxAngle = Double.NEGATIVE_INFINITY;
            for (DistanceLayer l : sorted) {
                double c = l.data[x];
                
                if (c >= maxAngle) {
                    maxAngle = c;
                } else {
                    l.data[x] = Double.NaN;
                }
            }
        }
        
        return sorted;
    }

    public static boolean isVisible(Collection<DistanceLayer> layers, int distance, int x, double visibleAngle) {
        List<DistanceLayer> sorted = layers.stream()
                .sorted()
                .collect(Collectors.toList());
        
        for (DistanceLayer l : sorted) {
            if (l.farDistance > (distance - 200)) {
                return true;
            }
            
            double c = l.data[x];

            if (c >= visibleAngle) {
                return false;
            }
        }
        
        return true;
    }

}
