package ch.inf3.horizoncut.data;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class Position {

    public final int gridLat;
    public final int gridLon;

    public final int row;
    public final int col;

    private final double lat;
    private final double lon;

    private static final Logger LOG = LoggerFactory.getLogger(Position.class);

    public Position(double lat, double lon) {
        this.gridLat = integerPart(lat);
        this.gridLon = integerPart(lon);
        this.row = (int) Math.floor(fractionalPart(lat) * Tile.RESOLUTION_SRTM1);
        this.col = (int) Math.floor(fractionalPart(lon) * Tile.RESOLUTION_SRTM1);

        this.lat = lat;
        this.lon = lon;
    }

    public static Position get(Position coordinates, double angleRad, int distance) {
        final double rEarth = 6371 * 1000;

        double rDistance = (distance * 1.0) / rEarth; // normalize linear distance to radian angle
        double φ1 = coordinates.getLatitudeRad();
        double λ1 = coordinates.getLongitudeRad();

        double φ2 = Math.asin(Math.sin(φ1) * Math.cos(rDistance)
                + Math.cos(φ1) * Math.sin(rDistance) * Math.cos(angleRad));
        double λ2 = λ1 + Math.atan2(Math.sin(angleRad) * Math.sin(rDistance) * Math.cos(φ1),
                Math.cos(rDistance) - Math.sin(φ1) * Math.sin(φ2));

        return new Position(Math.toDegrees(φ2), Math.toDegrees(λ2));
    }

    public int distanceToUnits(Position other) {
        int dxGrid = this.gridLat - other.gridLat;
        int dxCol = this.col - other.col;

        int dyGrid = this.gridLon - other.gridLon;
        int dyRow = this.row - other.row;

        int dx = dxGrid * Tile.RESOLUTION_SRTM1 + dxCol;
        int dy = dyGrid * Tile.RESOLUTION_SRTM1 + dyRow;

        return (int) Math.sqrt(dx * dx + dy * dy);
    }

    public int distanceToMeters(Position other) {
        final double R = 6371 * 1000;

        double dLat = Math.toRadians(other.getLatitude() - this.getLatitude());
        double dLon = Math.toRadians(other.getLongitude()- this.getLongitude());

        double a = Math.sin(dLat / 2) * Math.sin(dLat / 2)
                + Math.cos(this.getLatitude() * (Math.PI / 180.0)) * Math.cos(other.getLatitude() * (Math.PI / 180.0))
                * Math.sin(dLon / 2) * Math.sin(dLon / 2);

        double c = 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1 - a));
        int d = (int) (R * c);
        return d;
    }

    public static double bearing(Position a, Position b) {

        double φ1 = a.getLatitudeRad();
        double φ2 = b.getLatitudeRad();

        double λ1 = a.getLongitudeRad();
        double λ2 = b.getLongitudeRad();

        double y = Math.sin(λ2 - λ1) * Math.cos(φ2);
        double x = Math.cos(φ1) * Math.sin(φ2)
                - Math.sin(φ1) * Math.cos(φ2) * Math.cos(λ2 - λ1);
        double θ = Math.atan2(y, x);
        double brng = (θ + 2 * Math.PI) % (2 * Math.PI);

        return brng;
    }

    static int mod(int a, int b) {
        int mod = a % b;
        // We still need to move negative values into the range
        return (mod + b) % b;
    }

    static int integerPart(double d) {
        return (int) d;
    }

    static double fractionalPart(double d) {
        return d - integerPart(d);
    }

    public double getLatitude() {
        return this.lat;
    }

    public double getLongitude() {
        return this.lon;
    }

    public double getLatitudeRad() {
        return Math.toRadians(getLatitude());
    }

    public double getLongitudeRad() {
        return Math.toRadians(getLongitude());
    }

    public String getLatLon() {
        return String.format("(%.4f, %.4f)", getLatitude(), getLongitude());
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + gridLat;
        result = prime * result + gridLon;
        result = prime * result + col;
        result = prime * result + row;
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        Position other = (Position) obj;
        if (gridLat != other.gridLat) {
            return false;
        }
        if (gridLon != other.gridLon) {
            return false;
        }
        if (col != other.col) {
            return false;
        }
        if (row != other.row) {
            return false;
        }
        return true;
    }

    @Override
    public String toString() {
        return "Position[gridLat=" + gridLat + ", gridLon=" + gridLon + ", col=" + col + ", row=" + row + "]";
    }

}
