package ch.inf3.horizoncut.data;

import java.nio.ShortBuffer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Height data for a one by one degree latitude and longitude tile in high
 * resolution (E.g. N44E016), read from a HGT file.
 *
 * The tile is divided to 3600 rows and 3600 columns. The real-world size of
 * such pixel depends on its latitude and longitude. - Degrees of latitude are
 * parallel, so the distance between each degree is mostly constant, 1 degree ~=
 * 111km, so the height of each pixel is roughly 30.8 m. - Longitude varies
 * based on the actual location. It is also ~= 111km at the equator, but shrinks
 * to 0 as it reaches the poles.
 *
 * @see https://lpdaac.usgs.gov/documents/179/SRTM_User_Guide_V3.pdf
 */
public class Tile {
    public static final int RESOLUTION_SRTM1 = 3601;
    public static final int RESOLUTION_SRTM3 = 1201;

    private static final Logger LOG = LoggerFactory.getLogger(Tile.class);

    private final ShortBuffer data;

    private final int lon;

    private final int lat;

    private boolean isMaxElevationCalculated = false;
    private double maxElevation = 0;

    Tile(int lat, int lon, ShortBuffer data) {
        this.lat = lat;
        this.lon = lon;
        this.data = data;
    }

    public double elevation(final int row, final int col) {
        if (row < 0 || (RESOLUTION_SRTM1 - 1) < row || col < 0 || (RESOLUTION_SRTM1 - 1) < col) {
            LOG.error("{}x{} is not within the file for {},{}", row, col, lat, lon);
            throw new IllegalArgumentException(createExceptionMessage(row, col));
        }

        int cell = (RESOLUTION_SRTM1 * (RESOLUTION_SRTM1 - row - 1)) + col;

        if (data.limit() <= cell) {
            throw new IllegalArgumentException(createExceptionMessage(row, col));
        }

        short elevation;
        try {
            elevation = data.get(cell);
        } catch (IndexOutOfBoundsException e) {
            throw new IllegalArgumentException(createExceptionMessage(row, col), e);
        }
        if (elevation == Short.MIN_VALUE) {
            return Double.NaN;
        } else {
            return elevation;
        }
    }

    public double getMaxElevation() {
        if (!isMaxElevationCalculated) {
            isMaxElevationCalculated = true;
            for (int x = 0; x < RESOLUTION_SRTM1; x++) {
                for (int y = 0; y < RESOLUTION_SRTM1; y++) {
                    final double elevation = elevation(x, y);
                    if (elevation > maxElevation) {
                        maxElevation = elevation;
                    }
                }
            }
        }
        return maxElevation;
    }

    public double elevationByExactCoordinates(final double lat, final double lon) {
        if (lat < this.lat || lat >= this.lat + 1 || lon < this.lon || lon >= this.lon + 1) {
            throw new IllegalArgumentException("{lat: " + lat + " lon: " + lon + "} is not in tile of {lat: "
                    + this.lat + " lon: " + this.lon + "}");
        }
        double latDelta = lat - this.lat;
        double lonDelta = lon - this.lon;

        int row = (int) (RESOLUTION_SRTM1 * latDelta);
        int col = (int) (RESOLUTION_SRTM1 * lonDelta);

        return elevation(row, col);
    }

    private static String createExceptionMessage(int row, int col) {
        return "Query: " + row + "/" + col;
    }

    public int lat() {
        return lat;
    }

    public int lon() {
        return lon;
    }

    static class EmptyTile extends Tile {

        EmptyTile(int lat, int lon, ShortBuffer data) {
            super(lat, lon, data);

            System.out.println("Failed to access tile at lat/lon " + lat + "/" + lon + ". Using zero elevations in this region.");
        }

        @Override
        public double elevation(int row, int col) {
            return 0;
        }
    }

}
