package ch.inf3.horizoncut.data;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.channels.FileChannel;
import java.util.ArrayList;
import java.util.List;

/**
 * Read HGT files into {@link Tile} objects to form a {@link TileMap}.
 */
public class TileReader {

    private final String baseDir;

    public TileReader(String baseDir) {
        this.baseDir = baseDir;
    }

    public TileMap readHGTFiles(Position c, int radius) {
        List<Tile> tiles = new ArrayList<>();

        for (int lat = c.gridLat - radius; lat <= c.gridLat + radius; lat++) {
            for (int lon = c.gridLon - radius; lon <= c.gridLon + radius; lon++) {
                tiles.add(readHgtFile(lat, lon));
            }
        }

        return new TileMap(tiles);
    }

    private Tile readHgtFile(final int lat, final int lon) {
        File hgtFile = getHGTFile(lat, lon);

        try(FileInputStream fis = new FileInputStream(hgtFile);
                FileChannel fc = fis.getChannel()) {

            ByteBuffer bb = ByteBuffer.allocateDirect((int) fc.size());
            while (bb.remaining() > 0) {
                fc.read(bb);
            }

            bb.flip();
            return new Tile(lat, lon, bb.order(ByteOrder.BIG_ENDIAN).asShortBuffer());
        } catch (IOException e) {
            return new Tile.EmptyTile(lat, lon, null);
        }
    }

    private File getHGTFile(int lat, int lon) {
        char hWe;
        char hNs;

        if (lat < 0) {
            hNs = 'S';
        } else {
            hNs = 'N';
        }

        if (lon < 0) {
            hWe = 'W';
        } else {
            hWe = 'E';
        }

        String fileName = String.format("%c%02d%c%03d.hgt", hNs, lat, hWe, lon);

        return new File(baseDir, fileName);
    }
}
